'use strict';

var path = require('path');
var LRUCache = require('lru-cache');
var winston = require('winston');
var database = require('@nocobase/database');
var server = require('@nocobase/server');
var utils = require('@nocobase/utils');
var logger = require('@nocobase/logger');
var Processor = require('./Processor');
var initActions = require('./actions');
var constants = require('./constants');
var initFields = require('./fields');
var initFunctions = require('./functions');
var initInstructions = require('./instructions');
var initTriggers = require('./triggers');

function _interopDefault (e) { return e && e.__esModule ? e : { default: e }; }

var path__default = /*#__PURE__*/_interopDefault(path);
var LRUCache__default = /*#__PURE__*/_interopDefault(LRUCache);
var winston__default = /*#__PURE__*/_interopDefault(winston);
var Processor__default = /*#__PURE__*/_interopDefault(Processor);
var initActions__default = /*#__PURE__*/_interopDefault(initActions);
var initFields__default = /*#__PURE__*/_interopDefault(initFields);
var initFunctions__default = /*#__PURE__*/_interopDefault(initFunctions);
var initInstructions__default = /*#__PURE__*/_interopDefault(initInstructions);
var initTriggers__default = /*#__PURE__*/_interopDefault(initTriggers);

class WorkflowPlugin extends server.Plugin {
  instructions = new utils.Registry();
  triggers = new utils.Registry();
  functions = new utils.Registry();
  executing = false;
  pending = [];
  events = [];
  loggerCache;
  getLogger(workflowId) {
    const now = /* @__PURE__ */ new Date();
    const date = `${now.getFullYear()}-${`0${now.getMonth() + 1}`.slice(-2)}-${`0${now.getDate()}`.slice(-2)}`;
    const key = `${date}-${workflowId}}`;
    if (this.loggerCache.has(key)) {
      return this.loggerCache.get(key);
    }
    const logger$1 = logger.createLogger({
      transports: [
        "console",
        new winston__default.default.transports.File({
          filename: logger.getLoggerFilePath("workflows", date, `${workflowId}.log`),
          level: logger.getLoggerLevel()
        })
      ]
    });
    this.loggerCache.set(key, logger$1);
    return logger$1;
  }
  onBeforeSave = async (instance, options) => {
    const Model = instance.constructor;
    if (instance.enabled) {
      instance.set("current", true);
    } else if (!instance.current) {
      const count = await Model.count({
        where: {
          key: instance.key
        },
        transaction: options.transaction
      });
      if (!count) {
        instance.set("current", true);
      }
    }
    if (!instance.changed("enabled") || !instance.enabled) {
      return;
    }
    const previous = await Model.findOne({
      where: {
        key: instance.key,
        current: true,
        id: {
          [database.Op.ne]: instance.id
        }
      },
      transaction: options.transaction
    });
    if (previous) {
      await previous.update(
        { enabled: false, current: null },
        {
          transaction: options.transaction,
          hooks: false
        }
      );
      this.toggle(previous, false);
    }
  };
  async load() {
    const { db, options } = this;
    initFields__default.default(this);
    initActions__default.default(this);
    initTriggers__default.default(this, options.triggers);
    initInstructions__default.default(this, options.instructions);
    initFunctions__default.default(this, options.functions);
    this.loggerCache = new LRUCache__default.default({
      max: 20,
      updateAgeOnGet: true,
      dispose(logger) {
        logger.end();
      }
    });
    this.app.acl.registerSnippet({
      name: `pm.${this.name}.workflows`,
      actions: [
        "workflows:*",
        "workflows.nodes:*",
        "executions:list",
        "executions:get",
        "flow_nodes:update",
        "flow_nodes:destroy"
      ]
    });
    this.app.acl.registerSnippet({
      name: "ui.*",
      actions: ["workflows:list"]
    });
    this.app.acl.allow("users_jobs", ["list", "get", "submit"], "loggedIn");
    this.app.acl.allow("workflows", ["trigger"], "loggedIn");
    await db.import({
      directory: path__default.default.resolve(__dirname, "collections")
    });
    this.db.addMigrations({
      namespace: "workflow",
      directory: path__default.default.resolve(__dirname, "migrations"),
      context: {
        plugin: this
      }
    });
    db.on("workflows.beforeSave", this.onBeforeSave);
    db.on("workflows.afterSave", (model) => this.toggle(model));
    db.on("workflows.afterDestroy", (model) => this.toggle(model, false));
    this.app.on("beforeStart", async () => {
      const collection = db.getCollection("workflows");
      const workflows = await collection.repository.find({
        filter: { enabled: true }
      });
      workflows.forEach((workflow) => {
        this.toggle(workflow);
      });
    });
    this.app.on("afterStart", () => {
      this.app.setMaintainingMessage("check for not started executions");
      this.dispatch();
    });
    this.app.on("beforeStop", async () => {
      const collection = db.getCollection("workflows");
      const workflows = await collection.repository.find({
        filter: { enabled: true }
      });
      workflows.forEach((workflow) => {
        this.toggle(workflow, false);
      });
    });
  }
  toggle(workflow, enable) {
    const type = workflow.get("type");
    const trigger = this.triggers.get(type);
    if (typeof enable !== "undefined" ? enable : workflow.get("enabled")) {
      const prev = workflow.previous();
      if (prev.config) {
        trigger.off({ ...workflow.get(), ...prev });
      }
      trigger.on(workflow);
    } else {
      trigger.off(workflow);
    }
  }
  trigger(workflow, context, options = {}) {
    if (context == null) {
      return;
    }
    this.events.push([workflow, context, options]);
    this.getLogger(workflow.id).info(`new event triggered, now events: ${this.events.length}`);
    this.getLogger(workflow.id).debug(`event data:`, {
      data: context
    });
    if (this.events.length > 1) {
      return;
    }
    setTimeout(this.prepare);
  }
  async resume(job) {
    if (!job.execution) {
      job.execution = await job.getExecution();
    }
    this.pending.push([job.execution, job]);
    this.dispatch();
  }
  createProcessor(execution, options = {}) {
    return new Processor__default.default(execution, { ...options, plugin: this });
  }
  prepare = async () => {
    var _a;
    const [event] = this.events;
    if (!event) {
      return;
    }
    const [workflow, context, options] = event;
    let valid = true;
    if ((_a = options.context) == null ? void 0 : _a.executionId) {
      const existed = await workflow.countExecutions({
        where: {
          id: options.context.executionId
        }
      });
      if (existed) {
        this.getLogger(workflow.id).warn(
          `workflow ${workflow.id} has already been triggered in same execution (${options.context.executionId}), and newly triggering will be skipped.`
        );
        valid = false;
      }
    }
    if (valid) {
      const execution = await this.db.sequelize.transaction(async (transaction) => {
        const execution2 = await workflow.createExecution(
          {
            context,
            key: workflow.key,
            status: constants.EXECUTION_STATUS.QUEUEING
          },
          { transaction }
        );
        await workflow.increment("executed", { transaction });
        await workflow.constructor.increment("allExecuted", {
          where: {
            key: workflow.key
          },
          transaction
        });
        execution2.workflow = workflow;
        return execution2;
      });
      this.getLogger(workflow.id).debug(`execution of workflow ${workflow.id} created as ${execution.id}`, {
        data: execution.context
      });
      if (!this.executing && !this.pending.length) {
        this.pending.push([execution]);
      }
    }
    this.events.shift();
    if (this.events.length) {
      await this.prepare();
    } else {
      this.dispatch();
    }
  };
  async dispatch() {
    if (this.executing) {
      return;
    }
    this.executing = true;
    let next = null;
    if (this.pending.length) {
      next = this.pending.shift();
      this.getLogger(next[0].workflowId).info(`pending execution (${next[0].id}) ready to process`);
    } else {
      const execution = await this.db.getRepository("executions").findOne({
        filter: {
          status: constants.EXECUTION_STATUS.QUEUEING
        },
        appends: ["workflow"],
        sort: "createdAt"
      });
      if (execution && execution.workflow.enabled) {
        this.getLogger(execution.workflowId).info(`execution (${execution.id}) fetched from db`);
        next = [execution];
      }
    }
    if (next) {
      await this.process(...next);
    }
    this.executing = false;
    if (next) {
      this.dispatch();
    }
  }
  async process(execution, job) {
    var _a, _b;
    if (execution.status === constants.EXECUTION_STATUS.QUEUEING) {
      await execution.update({ status: constants.EXECUTION_STATUS.STARTED });
    }
    const processor = this.createProcessor(execution);
    this.getLogger(execution.workflowId).info(`execution (${execution.id}) ${job ? "resuming" : "starting"}...`);
    try {
      await (job ? processor.resume(job) : processor.start());
      this.getLogger(execution.workflowId).info(
        `execution (${execution.id}) finished with status: ${execution.status}`
      );
      if (execution.status && ((_b = (_a = execution.workflow.options) == null ? void 0 : _a.deleteExecutionOnStatus) == null ? void 0 : _b.includes(execution.status))) {
        await execution.destroy();
      }
    } catch (err) {
      this.getLogger(execution.workflowId).error(`execution (${execution.id}) error: ${err.message}`, err);
    }
  }
}

module.exports = WorkflowPlugin;
