'use strict';

var database = require('@nocobase/database');
var evaluators = require('@nocobase/evaluators');
var utils = require('@nocobase/utils');
var constants = require('./constants');

class Processor {
  constructor(execution, options) {
    this.execution = execution;
    this.options = options;
    this.logger = options.plugin.getLogger(execution.workflowId);
  }
  static StatusMap = {
    [constants.JOB_STATUS.PENDING]: constants.EXECUTION_STATUS.STARTED,
    [constants.JOB_STATUS.RESOLVED]: constants.EXECUTION_STATUS.RESOLVED,
    [constants.JOB_STATUS.FAILED]: constants.EXECUTION_STATUS.FAILED,
    [constants.JOB_STATUS.ERROR]: constants.EXECUTION_STATUS.ERROR,
    [constants.JOB_STATUS.ABORTED]: constants.EXECUTION_STATUS.ABORTED,
    [constants.JOB_STATUS.CANCELED]: constants.EXECUTION_STATUS.CANCELED,
    [constants.JOB_STATUS.REJECTED]: constants.EXECUTION_STATUS.REJECTED
  };
  logger;
  transaction;
  nodes = [];
  nodesMap = /* @__PURE__ */ new Map();
  jobsMap = /* @__PURE__ */ new Map();
  jobsMapByNodeId = {};
  // make dual linked nodes list then cache
  makeNodes(nodes = []) {
    this.nodes = nodes;
    nodes.forEach((node) => {
      this.nodesMap.set(node.id, node);
    });
    nodes.forEach((node) => {
      if (node.upstreamId) {
        node.upstream = this.nodesMap.get(node.upstreamId);
      }
      if (node.downstreamId) {
        node.downstream = this.nodesMap.get(node.downstreamId);
      }
    });
  }
  makeJobs(jobs) {
    jobs.forEach((job) => {
      this.jobsMap.set(job.id, job);
      this.jobsMapByNodeId[job.nodeId] = job.result;
    });
  }
  async getTransaction() {
    var _a;
    if (!((_a = this.execution.workflow.options) == null ? void 0 : _a.useTransaction)) {
      return;
    }
    const { options } = this;
    return options.transaction && !options.transaction.finished ? options.transaction : await options.plugin.db.sequelize.transaction();
  }
  async prepare() {
    const { execution } = this;
    if (!execution.workflow) {
      execution.workflow = await execution.getWorkflow();
    }
    const transaction = await this.getTransaction();
    this.transaction = transaction;
    const nodes = await execution.workflow.getNodes();
    this.makeNodes(nodes);
    const jobs = await execution.getJobs({
      order: [["id", "ASC"]],
      transaction
    });
    this.makeJobs(jobs);
  }
  async start() {
    const { execution } = this;
    if (execution.status !== constants.EXECUTION_STATUS.STARTED) {
      throw new Error(`execution was ended with status ${execution.status}`);
    }
    await this.prepare();
    if (this.nodes.length) {
      const head = this.nodes.find((item) => !item.upstream);
      await this.run(head, { result: execution.context });
    } else {
      await this.exit(constants.JOB_STATUS.RESOLVED);
    }
  }
  async resume(job) {
    const { execution } = this;
    if (execution.status !== constants.EXECUTION_STATUS.STARTED) {
      throw new Error(`execution was ended with status ${execution.status}`);
    }
    await this.prepare();
    const node = this.nodesMap.get(job.nodeId);
    await this.recall(node, job);
  }
  async commit() {
    if (this.transaction && (!this.options.transaction || this.options.transaction.finished)) {
      await this.transaction.commit();
    }
  }
  async exec(instruction, node, prevJob) {
    let job;
    try {
      this.logger.info(`execution (${this.execution.id}) run instruction [${node.type}] for node (${node.id})`);
      this.logger.debug(`config of node`, { data: node.config });
      job = await instruction(node, prevJob, this);
      if (!job) {
        return null;
      }
    } catch (err) {
      this.logger.error(
        `execution (${this.execution.id}) run instruction [${node.type}] for node (${node.id}) failed: `,
        { error: err }
      );
      job = {
        result: err instanceof Error ? { message: err.message, stack: process.env.NODE_ENV === "production" ? [] : err.stack } : err,
        status: constants.JOB_STATUS.ERROR
      };
      if (prevJob && prevJob.nodeId === node.id) {
        prevJob.set(job);
        job = prevJob;
      }
    }
    if (!(job instanceof database.Model)) {
      job.upstreamId = prevJob instanceof database.Model ? prevJob.get("id") : null;
      job.nodeId = node.id;
    }
    const savedJob = await this.saveJob(job);
    this.logger.info(
      `execution (${this.execution.id}) run instruction [${node.type}] for node (${node.id}) finished as status: ${savedJob.status}`
    );
    this.logger.debug(`result of node`, { data: savedJob.result });
    if (savedJob.status === constants.JOB_STATUS.RESOLVED && node.downstream) {
      this.logger.debug(`run next node (${node.downstreamId})`);
      return this.run(node.downstream, savedJob);
    }
    return this.end(node, savedJob);
  }
  async run(node, input) {
    const { instructions } = this.options.plugin;
    const instruction = instructions.get(node.type);
    if (typeof instruction.run !== "function") {
      return Promise.reject(new Error("`run` should be implemented for customized execution of the node"));
    }
    return this.exec(instruction.run.bind(instruction), node, input);
  }
  // parent node should take over the control
  async end(node, job) {
    this.logger.debug(`branch ended at node (${node.id})`);
    const parentNode = this.findBranchParentNode(node);
    if (parentNode) {
      this.logger.debug(`not on main, recall to parent entry node (${node.id})})`);
      await this.recall(parentNode, job);
      return job;
    }
    return this.exit(job.status);
  }
  async recall(node, job) {
    const { instructions } = this.options.plugin;
    const instruction = instructions.get(node.type);
    if (typeof instruction.resume !== "function") {
      return Promise.reject(new Error("`resume` should be implemented"));
    }
    return this.exec(instruction.resume.bind(instruction), node, job);
  }
  async exit(s) {
    if (typeof s === "number") {
      const status = this.constructor.StatusMap[s] ?? Math.sign(s);
      await this.execution.update({ status }, { transaction: this.transaction });
    }
    this.logger.info(`execution (${this.execution.id}) exiting with status ${this.execution.status}`);
    await this.commit();
    return null;
  }
  // TODO(optimize)
  async saveJob(payload) {
    const { database } = this.execution.constructor;
    const { model } = database.getCollection("jobs");
    let job;
    if (payload instanceof model) {
      job = await payload.save({ transaction: this.transaction });
    } else if (payload.id) {
      job = await model.findByPk(payload.id);
      await job.update(payload, {
        transaction: this.transaction
      });
    } else {
      job = await model.create(
        {
          ...payload,
          executionId: this.execution.id
        },
        {
          transaction: this.transaction
        }
      );
    }
    this.jobsMap.set(job.id, job);
    this.jobsMapByNodeId[job.nodeId] = job.result;
    return job;
  }
  getBranches(node) {
    return this.nodes.filter((item) => item.upstream === node && item.branchIndex !== null).sort((a, b) => Number(a.branchIndex) - Number(b.branchIndex));
  }
  // find the first node in current branch
  findBranchStartNode(node, parent) {
    for (let n = node; n; n = n.upstream) {
      if (!parent) {
        if (n.branchIndex !== null) {
          return n;
        }
      } else {
        if (n.upstream === parent) {
          return n;
        }
      }
    }
    return null;
  }
  // find the node start current branch
  findBranchParentNode(node) {
    for (let n = node; n; n = n.upstream) {
      if (n.branchIndex !== null) {
        return n.upstream;
      }
    }
    return null;
  }
  findBranchEndNode(node) {
    for (let n = node; n; n = n.downstream) {
      if (!n.downstream) {
        return n;
      }
    }
    return null;
  }
  findBranchParentJob(job, node) {
    for (let j = job; j; j = this.jobsMap.get(j.upstreamId)) {
      if (j.nodeId === node.id) {
        return j;
      }
    }
    return null;
  }
  findBranchLastJob(node) {
    for (let n = this.findBranchEndNode(node); n && n !== node.upstream; n = n.upstream) {
      const jobs = Array.from(this.jobsMap.values()).filter((item) => item.nodeId === n.id).sort((a, b) => a.createdAt.getTime() - b.createdAt.getTime());
      if (jobs.length) {
        return jobs[jobs.length - 1];
      }
    }
    return null;
  }
  getScope(sourceNodeId) {
    const node = this.nodesMap.get(sourceNodeId);
    const systemFns = {};
    const scope = {
      execution: this.execution,
      node
    };
    for (const [name, fn] of this.options.plugin.functions.getEntities()) {
      systemFns[name] = fn.bind(scope);
    }
    const $scopes = {};
    for (let n = this.findBranchParentNode(node); n; n = this.findBranchParentNode(n)) {
      const instruction = this.options.plugin.instructions.get(n.type);
      if (typeof instruction.getScope === "function") {
        $scopes[n.id] = instruction.getScope(n, this.jobsMapByNodeId[n.id], this);
      }
    }
    return {
      $context: this.execution.context,
      $jobsMapByNodeId: this.jobsMapByNodeId,
      $system: systemFns,
      $scopes
    };
  }
  getParsedValue(value, sourceNodeId, additionalScope) {
    const template = utils.parse(value);
    const scope = Object.assign(this.getScope(sourceNodeId), additionalScope);
    template.parameters.forEach(({ key }) => {
      evaluators.appendArrayColumn(scope, key);
    });
    return template(scope);
  }
}

module.exports = Processor;
