'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var constants = require('../constants');

const PARALLEL_MODE = {
  ALL: "all",
  ANY: "any",
  RACE: "race"
};
const Modes = {
  [PARALLEL_MODE.ALL]: {
    next(previous) {
      return previous.status >= constants.JOB_STATUS.PENDING;
    },
    getStatus(result) {
      const failedStatus = result.find((status) => status != null && status < constants.JOB_STATUS.PENDING);
      if (typeof failedStatus !== "undefined") {
        return failedStatus;
      }
      if (result.every((status) => status != null && status === constants.JOB_STATUS.RESOLVED)) {
        return constants.JOB_STATUS.RESOLVED;
      }
      return constants.JOB_STATUS.PENDING;
    }
  },
  [PARALLEL_MODE.ANY]: {
    next(previous) {
      return previous.status <= constants.JOB_STATUS.PENDING;
    },
    getStatus(result) {
      if (result.some((status) => status != null && status === constants.JOB_STATUS.RESOLVED)) {
        return constants.JOB_STATUS.RESOLVED;
      }
      if (result.some((status) => status != null ? status === constants.JOB_STATUS.PENDING : true)) {
        return constants.JOB_STATUS.PENDING;
      }
      return constants.JOB_STATUS.FAILED;
    }
  },
  [PARALLEL_MODE.RACE]: {
    next(previous) {
      return previous.status === constants.JOB_STATUS.PENDING;
    },
    getStatus(result) {
      if (result.some((status) => status != null && status === constants.JOB_STATUS.RESOLVED)) {
        return constants.JOB_STATUS.RESOLVED;
      }
      const failedStatus = result.find((status) => status != null && status < constants.JOB_STATUS.PENDING);
      if (typeof failedStatus !== "undefined") {
        return failedStatus;
      }
      return constants.JOB_STATUS.PENDING;
    }
  }
};
var parallel_default = {
  async run(node, prevJob, processor) {
    const branches = processor.getBranches(node);
    const job = await processor.saveJob({
      status: constants.JOB_STATUS.PENDING,
      result: Array(branches.length).fill(null),
      nodeId: node.id,
      upstreamId: (prevJob == null ? void 0 : prevJob.id) ?? null
    });
    const { mode = PARALLEL_MODE.ALL } = node.config;
    await branches.reduce(
      (promise, branch, i) => promise.then(async (previous) => {
        if (i && !Modes[mode].next(previous)) {
          return previous;
        }
        await processor.run(branch, job);
        return processor.findBranchLastJob(branch);
      }),
      Promise.resolve()
    );
    return null;
  },
  async resume(node, branchJob, processor) {
    const job = processor.findBranchParentJob(branchJob, node);
    const { result, status } = job;
    if (status !== constants.JOB_STATUS.PENDING) {
      return processor.exit();
    }
    const jobNode = processor.nodesMap.get(branchJob.nodeId);
    const branchStartNode = processor.findBranchStartNode(jobNode, node);
    const branches = processor.getBranches(node);
    const branchIndex = branches.indexOf(branchStartNode);
    const { mode = PARALLEL_MODE.ALL } = node.config || {};
    const newResult = [...result.slice(0, branchIndex), branchJob.status, ...result.slice(branchIndex + 1)];
    job.set({
      result: newResult,
      status: Modes[mode].getStatus(newResult)
    });
    if (job.status === constants.JOB_STATUS.PENDING) {
      await job.save({ transaction: processor.transaction });
      return processor.exit();
    }
    return job;
  }
};

exports.PARALLEL_MODE = PARALLEL_MODE;
exports.default = parallel_default;
